; #################################################### Initstuff #################################################

; #################################################### Includes ##################################################

; #################################################### Prototypes ################################################

; #################################################### Structures ################################################

Structure Data_Source
  Type.i
  
  Name.s
  File_ID.i
  hProcess.i
  
  Resizeable.i
  Writeable.i
EndStructure

; #################################################### Variables #################################################

; #################################################### Constants #################################################

Enumeration
  #Data_Source_Debug
  #Data_Source_Random
  
  #Data_Source_File
  #Data_Source_Process
  #Data_Source_Clipboard
EndEnumeration

; #################################################### Declares ##################################################

; #################################################### Macros ####################################################

; #################################################### Procedures ################################################

Procedure Data_Source_Destroy(*Data_Source.Data_Source)
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  
  If *Data_Source\File_ID
    CloseFile(*Data_Source\File_ID)
  EndIf
  
  If *Data_Source\hProcess
    CloseHandle_(*Data_Source\hProcess)
  EndIf
    
  ClearStructure(*Data_Source, Data_Source)
  FreeMemory(*Data_Source)
  
  ProcedureReturn #True
EndProcedure

Procedure Data_Source_Create(Type, Name.s, Create=0)
  Protected *Data_Source.Data_Source = AllocateMemory(SizeOf(Data_Source))
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  InitializeStructure(*Data_Source, Data_Source)
  
  Select Type
    Case #Data_Source_Debug
      *Data_Source\Name = Name
      *Data_Source\Type = Type
      *Data_Source\Resizeable = #False
      
    Case #Data_Source_Random
      *Data_Source\Name = Name
      *Data_Source\Type = Type
      *Data_Source\Resizeable = #False
      
    Case #Data_Source_File
      *Data_Source\Type = Type
      *Data_Source\Name = Name
      *Data_Source\Resizeable = #True
      If FileSize(Name) = -1 Or Create
        *Data_Source\File_ID = CreateFile(#PB_Any, Name)
        If Not *Data_Source\File_ID
          MessageRequester("D3hex", "Couldn't create the file.")
          Data_Source_Destroy(*Data_Source)
          ProcedureReturn #False
        EndIf
        *Data_Source\Writeable = 1
      Else
        *Data_Source\File_ID = ReadFile(#PB_Any, Name)
        If Not *Data_Source\File_ID
          MessageRequester("D3hex", "Couldn't open the file.")
          Data_Source_Destroy(*Data_Source)
          ProcedureReturn #False
        EndIf
      EndIf
      ;FileBuffersSize(*Data_Source\File_ID, 0)
      
    Case #Data_Source_Process
      *Data_Source\Type = Type
      *Data_Source\Name = Name
      *Data_Source\Resizeable = #False
      *Data_Source\hProcess = OpenProcess_(#MAXIMUM_ALLOWED, #False, Val(Name))
      If Not *Data_Source\hProcess
        MessageRequester("D3hex", "Couldn't open the process. ("+Str(GetLastError_())+")")
        Data_Source_Destroy(*Data_Source)
        ProcedureReturn #False
      EndIf
      
    Case #Data_Source_Clipboard
      *Data_Source\Name = Name
      *Data_Source\Type = Type
      *Data_Source\Resizeable = #True
      
    Default
      Data_Source_Destroy(*Data_Source)
      ProcedureReturn #False
  EndSelect
  
  ProcedureReturn *Data_Source
EndProcedure

Procedure.q Data_Source_Get_Size(*Data_Source.Data_Source)
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  
  Select *Data_Source\Type
    Case #Data_Source_Debug
      ProcedureReturn $000100000000;0000
      
    Case #Data_Source_Random
      ProcedureReturn $000100000000;0000
      
    Case #Data_Source_File
      ProcedureReturn Lof(*Data_Source\File_ID)
      
    Case #Data_Source_Process
      Protected MBI.MEMORY_BASIC_INFORMATION
      Protected Offset.q, Temp_Size.q, Last_BaseAdress.q
      While VirtualQueryEx_(*Data_Source\hProcess, Offset, MBI, SizeOf(MEMORY_BASIC_INFORMATION))
        If Last_BaseAdress > MBI\BaseAddress
          Break
        EndIf
        Select MBI\State
          Case #MEM_RESERVE, #MEM_COMMIT
            Temp_Size = MBI\BaseAddress + MBI\RegionSize
        EndSelect
        Offset = MBI\BaseAddress + MBI\RegionSize
        Last_BaseAdress = MBI\BaseAddress
      Wend
      ProcedureReturn Temp_Size
      
    Case #Data_Source_Clipboard
      Text.s = GetClipboardText()
      ProcedureReturn StringByteLength(Text)
      
  EndSelect
EndProcedure

Procedure Data_Source_Get_Ranges(*Data_Source.Data_Source, *Object.Object)
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  
  If Not *Object
    ProcedureReturn #False
  EndIf
  
  ClearList(*Object\Range())
  
  Select *Data_Source\Type
    Case #Data_Source_Debug
      AddElement(*Object\Range())
      *Object\Range()\Offset = 0
      *Object\Range()\Size = Data_Source_Get_Size(*Data_Source)
      *Object\Range()\Metadata = #Object_Metadata_Readable
      ProcedureReturn #True
      
    Case #Data_Source_Random
      AddElement(*Object\Range())
      *Object\Range()\Offset = 0
      *Object\Range()\Size = Data_Source_Get_Size(*Data_Source)
      *Object\Range()\Metadata = #Object_Metadata_Readable
      ProcedureReturn #True
      
    Case #Data_Source_File
      AddElement(*Object\Range())
      *Object\Range()\Offset = 0
      *Object\Range()\Size = Data_Source_Get_Size(*Data_Source)
      *Object\Range()\Metadata = #Object_Metadata_Readable | #Object_Metadata_Writeable
      ProcedureReturn #True
      
    Case #Data_Source_Process
      Protected MBI.MEMORY_BASIC_INFORMATION
      Protected Offset.q
      While VirtualQueryEx_(*Data_Source\hProcess, Offset, MBI, SizeOf(MEMORY_BASIC_INFORMATION))
        If LastElement(*Object\Range()) And *Object\Range()\Offset >= MBI\BaseAddress
          Break
        EndIf
        AddElement(*Object\Range())
        *Object\Range()\Offset = MBI\BaseAddress
        *Object\Range()\Size = MBI\RegionSize
        Select MBI\Protect & $FF
          Case #PAGE_NOACCESS           : *Object\Range()\Metadata = 0
          Case #PAGE_READONLY           : *Object\Range()\Metadata = #Object_Metadata_Readable
          Case #PAGE_READWRITE          : *Object\Range()\Metadata = #Object_Metadata_Readable | #Object_Metadata_Writeable
          Case #PAGE_WRITECOPY          : *Object\Range()\Metadata = #Object_Metadata_Readable | #Object_Metadata_Writeable
          Case #PAGE_EXECUTE            : *Object\Range()\Metadata = #Object_Metadata_Executable
          Case #PAGE_EXECUTE_READ       : *Object\Range()\Metadata = #Object_Metadata_Executable | #Object_Metadata_Readable
          Case #PAGE_EXECUTE_READWRITE  : *Object\Range()\Metadata = #Object_Metadata_Executable | #Object_Metadata_Readable | #Object_Metadata_Writeable
          Case #PAGE_EXECUTE_WRITECOPY  : *Object\Range()\Metadata = #Object_Metadata_Executable | #Object_Metadata_Readable | #Object_Metadata_Writeable
        EndSelect
        *Object\Range()\Type_State = MBI\lType | MBI\State
        Offset + MBI\RegionSize
      Wend
      ProcedureReturn #True
      
    Case #Data_Source_Clipboard
      AddElement(*Object\Range())
      *Object\Range()\Offset = 0
      *Object\Range()\Size = Data_Source_Get_Size(*Data_Source)
      *Object\Range()\Metadata = #Object_Metadata_Readable | #Object_Metadata_Writeable
      ProcedureReturn #True
      
  EndSelect
EndProcedure

Procedure Data_Source_Get(*Data_Source.Data_Source, Offset.q, Size.q, *Destination, *Metadata)
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  
  If Not *Destination
    ProcedureReturn #False
  EndIf
  
  If Offset < 0
    ProcedureReturn #False
  EndIf
  
  If Size <= 0
    ProcedureReturn #False
  EndIf
  
  FillMemory(*Metadata, Size, #Object_Metadata_Error)
  
  Select *Data_Source\Type
    Case #Data_Source_Debug
      FillMemory(*Metadata, Size, #Object_Metadata_Readable)
      For i = 0 To Size - 1
        PokeA(*Destination+i, Offset + i)
      Next
      ProcedureReturn #True
      
    Case #Data_Source_Random
      FillMemory(*Metadata, Size, #Object_Metadata_Readable)
      For i = 0 To Size-1
        PokeA(*Destination+i, CRC32Fingerprint(@Offset, SizeOf(Integer)))
        Offset + 1
      Next
      ProcedureReturn #True
      
    Case #Data_Source_File
      If Not *Data_Source\File_ID
        ProcedureReturn #False
      EndIf
      FillMemory(*Metadata, Size, #Object_Metadata_Readable)
      FileSeek(*Data_Source\File_ID, Offset)
      If Not ReadData(*Data_Source\File_ID, *Destination, Size)
        ProcedureReturn #False
      EndIf
      ProcedureReturn #True
      
    Case #Data_Source_Process
      Protected MBI.MEMORY_BASIC_INFORMATION
      FillMemory(*Destination, Size)
      
      While VirtualQueryEx_(*Data_Source\hProcess, Offset, MBI, SizeOf(MEMORY_BASIC_INFORMATION))
        Temp_Size.q = Size
        If Temp_Size >= MBI\BaseAddress + MBI\RegionSize - Offset : Temp_Size = MBI\BaseAddress + MBI\RegionSize - Offset : EndIf
        
        Select MBI\Protect & $FF
          Case #PAGE_NOACCESS           : FillMemory(*Metadata, Temp_Size)
          Case #PAGE_READONLY           : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Readable)
          Case #PAGE_READWRITE          : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Readable | #Object_Metadata_Writeable)
          Case #PAGE_WRITECOPY          : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Readable | #Object_Metadata_Writeable)
          Case #PAGE_EXECUTE            : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Executable)
          Case #PAGE_EXECUTE_READ       : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Executable | #Object_Metadata_Readable)
          Case #PAGE_EXECUTE_READWRITE  : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Executable | #Object_Metadata_Readable | #Object_Metadata_Writeable)
          Case #PAGE_EXECUTE_WRITECOPY  : FillMemory(*Metadata, Temp_Size, #Object_Metadata_Executable | #Object_Metadata_Readable | #Object_Metadata_Writeable)
        EndSelect
        
        Select MBI\State
          Case #MEM_FREE
            
          Case #MEM_RESERVE
            If Not ReadProcessMemory_(*Data_Source\hProcess, Offset, *Destination, Temp_Size, 0)
              ;FillMemory(*Metadata+Dest_Offset, Temp_Size, #Object_Metadata_Error)
            EndIf
          Case #MEM_COMMIT
            If Not ReadProcessMemory_(*Data_Source\hProcess, Offset, *Destination, Temp_Size, 0)
              ;FillMemory(*Metadata+Dest_Offset, Temp_Size, #Object_Metadata_Error)
            EndIf
        EndSelect
        
        Size - Temp_Size
        If Size > 0
          Offset + Temp_Size
          *Destination + Temp_Size
          *Metadata + Temp_Size
        Else
          Break
        EndIf
      Wend
      ProcedureReturn #True
      
    Case #Data_Source_Clipboard
      FillMemory(*Metadata, Size, #Object_Metadata_Readable | #Object_Metadata_Writeable)
      FillMemory(*Destination, Size)
      Text.s = GetClipboardText()
      Length.q = StringByteLength(Text)
      If Size > Length-Offset : Size = Length-Offset : EndIf
      If Size > 0
        CopyMemory(@Text+Offset, *Destination, Size)
      EndIf
      ProcedureReturn #True
      
  EndSelect
EndProcedure

Procedure Data_Source_Set_Writeable(*Data_Source.Data_Source, State)
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  
  If *Data_Source\Writeable = State
    ProcedureReturn #True
  EndIf
  
  Select *Data_Source\Type
    Case #Data_Source_Debug
      ProcedureReturn #False
      
    Case #Data_Source_Random
      ProcedureReturn #False
      
  EndSelect
  
  If State = 1
    If MessageRequester("D3hex", "Do you wanna switch to writing mode?", #PB_MessageRequester_YesNo) = #PB_MessageRequester_No
      ProcedureReturn #False
    EndIf
  EndIf
  
  Select *Data_Source\Type
    Case #Data_Source_File
      CloseFile(*Data_Source\File_ID)
      Select State
        Case 0
          *Data_Source\File_ID = ReadFile(#PB_Any, *Data_Source\Name)
          If Not *Data_Source\File_ID
            ProcedureReturn #False
          EndIf
          *Data_Source\Writeable = State
        Case 1
          *Data_Source\File_ID = OpenFile(#PB_Any, *Data_Source\Name)
          If Not *Data_Source\File_ID
            *Data_Source\File_ID = ReadFile(#PB_Any, *Data_Source\Name)
            *Data_Source\Writeable = 0
            ProcedureReturn #False
          EndIf
          If Not *Data_Source\File_ID
            ProcedureReturn #False
          EndIf
          *Data_Source\Writeable = State
      EndSelect
      ProcedureReturn #True
      
    Case #Data_Source_Process
      *Data_Source\Writeable = State
      ProcedureReturn #True
      
    Case #Data_Source_Clipboard
      *Data_Source\Writeable = State
      ProcedureReturn #True
      
  EndSelect
EndProcedure

Procedure Data_Source_Set(*Data_Source.Data_Source, Offset.q, Size.q, *Source)
  If Not *Data_Source
    ProcedureReturn #False
  EndIf
  
  If Not *Data_Source\Writeable
    ProcedureReturn #False
  EndIf
  
  If Not *Source
    ProcedureReturn #False
  EndIf
  
  If Offset < 0
    ProcedureReturn #False
  EndIf
  
  If Size <= 0
    ProcedureReturn #False
  EndIf
  
  Select *Data_Source\Type
    Case #Data_Source_Debug
      ProcedureReturn #False
      
    Case #Data_Source_Random
      ProcedureReturn #False
      
    Case #Data_Source_File
      If Not *Data_Source\File_ID
        ProcedureReturn #False
      EndIf
      FileSeek(*Data_Source\File_ID, Offset)
      If Not WriteData(*Data_Source\File_ID, *Source, Size)
        ProcedureReturn #False
      EndIf
      ProcedureReturn #True
      
    Case #Data_Source_Process
      Protected MBI.MEMORY_BASIC_INFORMATION
      
      Result = #True
      
      While VirtualQueryEx_(*Data_Source\hProcess, Offset, MBI, SizeOf(MEMORY_BASIC_INFORMATION))
        Temp_Size.q = Size
        If Temp_Size >= MBI\BaseAddress + MBI\RegionSize - Offset : Temp_Size = MBI\BaseAddress + MBI\RegionSize - Offset : EndIf
        
        Select MBI\Protect & $FF
          Case #PAGE_NOACCESS           : Result = #False
          Case #PAGE_READONLY           : Result = #False
          Case #PAGE_READWRITE          : 
          Case #PAGE_WRITECOPY          : 
          Case #PAGE_EXECUTE            : Result = #False
          Case #PAGE_EXECUTE_READ       : Result = #False
          Case #PAGE_EXECUTE_READWRITE  : 
          Case #PAGE_EXECUTE_WRITECOPY  : 
        EndSelect
        
        Select MBI\State
          Case #MEM_FREE
            Result = #False
          Case #MEM_RESERVE
            If Not WriteProcessMemory_(*Data_Source\hProcess, Offset, *Source, Temp_Size, 0)
              Result = #False
            EndIf
          Case #MEM_COMMIT
            If Not WriteProcessMemory_(*Data_Source\hProcess, Offset, *Source, Temp_Size, 0)
              Result = #False
            EndIf
        EndSelect
        
        Size - Temp_Size
        If Size > 0
          Offset + Temp_Size
          *Source + Temp_Size
        Else
          Break
        EndIf
      Wend
      
      If Size
        ProcedureReturn #False
      Else
        ProcedureReturn Result
      EndIf
      
    Case #Data_Source_Clipboard
      Length = Offset + Size
      Text.s = GetClipboardText()
      If Length < StringByteLength(Text) : Length = StringByteLength(Text) : EndIf
      *Temp = AllocateMemory(Length + 2)
      If Not *Temp
        ProcedureReturn #False
      EndIf
      
      CopyMemory(@Text, *Temp, StringByteLength(Text))
      CopyMemory(*Source, *Temp+Offset, Size)
      
      Text.s = PeekS(*Temp, Length)
      
      SetClipboardText(Text)
      
      FreeMemory(*Temp)
      ProcedureReturn #True
      
  EndSelect
EndProcedure

; #################################################### Initstuff #################################################

; #################################################### Datasections ##############################################
; IDE Options = PureBasic 4.60 Beta 1 (Windows - x64)
; CursorPosition = 131
; FirstLine = 115
; Folding = --
; EnableXP
; DisableDebugger