
; ############################################ Variablen ####################################################

Structure Memory_Operation
  Src_Offset.q
  Src_Size.q
  
  Dst_Offset.q
  Dst_Size.q
  
  Copy_Size.q
EndStructure

; ############################################ Proceduren ####################################################

; #### Cuts the Offset's / Sizes of the memory operation to prevent memory violations
Procedure Memory_Operation_Check(*Memory_Operation.Memory_Operation)
  Protected Temp.q
  
  If *Memory_Operation\Src_Offset < 0
    *Memory_Operation\Copy_Size  + *Memory_Operation\Src_Offset
    *Memory_Operation\Dst_Offset - *Memory_Operation\Src_Offset
    *Memory_Operation\Src_Offset - *Memory_Operation\Src_Offset
  EndIf
  
  If *Memory_Operation\Dst_Offset < 0
    *Memory_Operation\Copy_Size  + *Memory_Operation\Dst_Offset
    *Memory_Operation\Src_Offset - *Memory_Operation\Dst_Offset
    *Memory_Operation\Dst_Offset - *Memory_Operation\Dst_Offset
  EndIf
  
  Temp = *Memory_Operation\Src_Size - *Memory_Operation\Src_Offset
  If *Memory_Operation\Copy_Size > Temp
    *Memory_Operation\Copy_Size = Temp
  EndIf
  
  Temp = *Memory_Operation\Dst_Size - *Memory_Operation\Dst_Offset
  If *Memory_Operation\Copy_Size > Temp
    *Memory_Operation\Copy_Size = Temp
  EndIf
  
  If *Memory_Operation\Copy_Size < 0
    *Memory_Operation\Copy_Size = 0
  EndIf
  
  ProcedureReturn #True
EndProcedure

; #### Copies a specified amount of data (Copy_Size) from the source to the destination.
; #### It cuts everything, to prevent memory violations
Procedure Memory_Range_Copy(*Src, Src_Offset.q, *Dst, Dst_Offset.q, Copy_Size.q)
  If Not *Src
    ProcedureReturn #False
  EndIf
  
  If Not *Dst
    ProcedureReturn #False
  EndIf
  
  Protected Src_Size.q = MemorySize(*Src)
  Protected Dst_Size.q = MemorySize(*Dst)
  Protected Temp.q
  
  If Src_Offset < 0
    Copy_Size  + Src_Offset
    Dst_Offset - Src_Offset
    Src_Offset - Src_Offset
  EndIf
  
  If Dst_Offset < 0
    Copy_Size  + Dst_Offset
    Src_Offset - Dst_Offset
    Dst_Offset - Dst_Offset
  EndIf
  
  Temp = Src_Size - Src_Offset
  If Copy_Size > Temp
    Copy_Size = Temp
  EndIf
  
  Temp = Dst_Size - Dst_Offset
  If Copy_Size > Temp
    Copy_Size = Temp
  EndIf
  
  If Copy_Size > 0
    CopyMemory(*Src+Src_Offset, *Dst+Dst_Offset, Copy_Size)
  EndIf
  
  ProcedureReturn #True
EndProcedure

; #### Copies (MoveMemory) a specified amount of data (Copy_Size) from the source to the destination.
; #### It cuts everything, to prevent memory violations
Procedure Memory_Range_Move(*Src, Src_Offset.q, *Dst, Dst_Offset.q, Copy_Size.q)
  If Not *Src
    ProcedureReturn #False
  EndIf
  
  If Not *Dst
    ProcedureReturn #False
  EndIf
  
  Protected Src_Size.q = MemorySize(*Src)
  Protected Dst_Size.q = MemorySize(*Dst)
  Protected Temp.q
  
  If Src_Offset < 0
    Copy_Size  + Src_Offset
    Dst_Offset - Src_Offset
    Src_Offset - Src_Offset
  EndIf
  
  If Dst_Offset < 0
    Copy_Size  + Dst_Offset
    Src_Offset - Dst_Offset
    Dst_Offset - Dst_Offset
  EndIf
  
  Temp = Src_Size - Src_Offset
  If Copy_Size > Temp
    Copy_Size = Temp
  EndIf
  
  Temp = Dst_Size - Dst_Offset
  If Copy_Size > Temp
    Copy_Size = Temp
  EndIf
  
  If Copy_Size > 0
    MoveMemory(*Src+Src_Offset, *Dst+Dst_Offset, Copy_Size)
  EndIf
  
  ProcedureReturn #True
EndProcedure
; IDE Options = PureBasic 4.60 Beta 1 (Windows - x64)
; CursorPosition = 111
; Folding = -
; EnableXP
; DisableDebugger