
; ############################################ Variablen ####################################################

Structure Object_Hilbert_Main
  
EndStructure
Global Object_Hilbert_Main.Object_Hilbert_Main

Structure Object_Hilbert
  *Object.Object
  
  Window_ID.i
  Canvas.i
  ScrollArea.i
  
  Width.i
  Height.i
  Zoom_Factor.i
  Square_Size.i
  
  Redo_Timer.i
  Redo.i
  
  Name.s
EndStructure

Structure Hilbert_State
  X.i
  Y.i
  Size_X.i
  Size_Y.i
  
  *Object.Object
  Step_Size.q
  Offset.q
  Color.i
  Steps_To_Next_Color.q
EndStructure

Enumeration
  #Left
  #Right
  #Up
  #Down
EndEnumeration

#Object_Hilbert_Color_None            = $FFFFFF
#Object_Hilbert_Color_Free            = $000000
#Object_Hilbert_Color_Commit_Image    = $0000FF
#Object_Hilbert_Color_Commit_Mapped   = $00FF00
#Object_Hilbert_Color_Commit_Private  = $00FFFF
#Object_Hilbert_Color_Reserved        = $FFFF00

; ############################################ Proceduren ####################################################

Procedure Hilbert_Select_Color(*Hilbert_State.Hilbert_State)
  Found = 0
  
  While NextElement(*Hilbert_State\Object\Range())
    If *Hilbert_State\Object\Range()\Offset + *Hilbert_State\Object\Range()\Size > *Hilbert_State\Offset
      
      Found = 1
      
      *Hilbert_State\Steps_To_Next_Color = *Hilbert_State\Object\Range()\Size
      If *Hilbert_State\Object\Range()\Type_State & #MEM_FREE
        ; #### Memory is free
        *Hilbert_State\Color = #Object_Hilbert_Color_Free
      ElseIf *Hilbert_State\Object\Range()\Type_State & #MEM_COMMIT
        If *Hilbert_State\Object\Range()\Type_State & #MEM_IMAGE
          ; #### Memory is committed (Image)
          *Hilbert_State\Color = #Object_Hilbert_Color_Commit_Image
        ElseIf *Hilbert_State\Object\Range()\Type_State & #MEM_MAPPED
          ; #### Memory is committed (Mapped)
          *Hilbert_State\Color = #Object_Hilbert_Color_Commit_Mapped
        ElseIf *Hilbert_State\Object\Range()\Type_State & #MEM_PRIVATE
          ; #### Memory is committed (Private)
          *Hilbert_State\Color = #Object_Hilbert_Color_Commit_Private
        EndIf
      ElseIf *Hilbert_State\Object\Range()\Type_State & #MEM_RESERVE
        ; #### Memory is reserved
        *Hilbert_State\Color = #Object_Hilbert_Color_Reserved
      EndIf
      
      Break
    EndIf
  Wend
  
  If Found = 0
    *Hilbert_State\Color = #Object_Hilbert_Color_None
    *Hilbert_State\Steps_To_Next_Color = $FFFFFFFF
  EndIf
EndProcedure

Macro Hilbert_Move(Direction)
  *Hilbert_State\Offset + *Hilbert_State\Step_Size
  *Hilbert_State\Steps_To_Next_Color - *Hilbert_State\Step_Size
  If *Hilbert_State\Steps_To_Next_Color <= 0 : Hilbert_Select_Color(*Hilbert_State) : EndIf
  
  Select Direction
    Case #Left
      *Hilbert_State\X - 1
      If *Hilbert_State\X < *Hilbert_State\Size_X And *Hilbert_State\Y < *Hilbert_State\Size_Y
        Plot(*Hilbert_State\X, *Hilbert_State\Y, *Hilbert_State\Color)
      EndIf
    Case #Right
      *Hilbert_State\X + 1
      If *Hilbert_State\X < *Hilbert_State\Size_X And *Hilbert_State\Y < *Hilbert_State\Size_Y
        Plot(*Hilbert_State\X, *Hilbert_State\Y, *Hilbert_State\Color)
      EndIf
    Case #Up
      *Hilbert_State\Y - 1
      If *Hilbert_State\X < *Hilbert_State\Size_X And *Hilbert_State\Y < *Hilbert_State\Size_Y
        Plot(*Hilbert_State\X, *Hilbert_State\Y, *Hilbert_State\Color)
      EndIf
    Case #Down
      *Hilbert_State\Y + 1
      If *Hilbert_State\X < *Hilbert_State\Size_X And *Hilbert_State\Y < *Hilbert_State\Size_Y
        Plot(*Hilbert_State\X, *Hilbert_State\Y, *Hilbert_State\Color)
      EndIf
  EndSelect
EndMacro

Procedure Hilbert_Level(Level, *Hilbert_State.Hilbert_State, Direction=#Up)
  If Level = 1
    Select Direction
      Case #Left
        Hilbert_Move(#Right)
        Hilbert_Move(#Down)
        Hilbert_Move(#Left)
      Case #Right
        Hilbert_Move(#Left)
        Hilbert_Move(#Up)
        Hilbert_Move(#Right)
      Case #Up
        Hilbert_Move(#Down)
        Hilbert_Move(#Right)
        Hilbert_Move(#Up)
      Case #Down
        Hilbert_Move(#Up)
        Hilbert_Move(#Left)
        Hilbert_Move(#Down)
    EndSelect
  Else
    Select Direction
      Case #Left
        Hilbert_Level(Level-1, *Hilbert_State,#Up)
        Hilbert_Move(#Right)
        Hilbert_Level(Level-1, *Hilbert_State,#Left)
        Hilbert_Move(#Down)
        Hilbert_Level(Level-1, *Hilbert_State,#Left)
        Hilbert_Move(#Left)
        Hilbert_Level(Level-1, *Hilbert_State,#Down)
      Case #Right
        Hilbert_Level(Level-1, *Hilbert_State,#Down)
        Hilbert_Move(#Left)
        Hilbert_Level(Level-1, *Hilbert_State,#Right)
        Hilbert_Move(#Up)
        Hilbert_Level(Level-1, *Hilbert_State,#Right)
        Hilbert_Move(#Right)
        Hilbert_Level(Level-1, *Hilbert_State,#Up)
      Case #Up
        Hilbert_Level(Level-1, *Hilbert_State,#Left)
        Hilbert_Move(#Down)
        Hilbert_Level(Level-1, *Hilbert_State,#Up)
        Hilbert_Move(#Right)
        Hilbert_Level(Level-1, *Hilbert_State,#Up)
        Hilbert_Move(#Up)
        Hilbert_Level(Level-1, *Hilbert_State,#Right)
      Case #Down
        Hilbert_Level(Level-1, *Hilbert_State,#Right)
        Hilbert_Move(#Up)
        Hilbert_Level(Level-1, *Hilbert_State,#Down)
        Hilbert_Move(#Left)
        Hilbert_Level(Level-1, *Hilbert_State,#Down)
        Hilbert_Move(#Down)
        Hilbert_Level(Level-1, *Hilbert_State,#Left)
    EndSelect
  EndIf
EndProcedure

Procedure Object_Hilbert_View_Redo(*Object_Hilbert.Object_Hilbert, Width, Height)
  If Not *Object_Hilbert
    ProcedureReturn #False
  EndIf
  
  If Width > 0 And Height > 0
    
    If *Object_Hilbert\Width <> Width Or *Object_Hilbert\Height <> Height
      *Object_Hilbert\Width = Width
      *Object_Hilbert\Height = Height
      
      *Object_Hilbert\Square_Size = *Object_Hilbert\Width
      
      ResizeGadget(*Object_Hilbert\ScrollArea, 0, 0, *Object_Hilbert\Width, *Object_Hilbert\Height)
      SetGadgetAttribute(*Object_Hilbert\ScrollArea, #PB_ScrollArea_InnerWidth, *Object_Hilbert\Square_Size)
      SetGadgetAttribute(*Object_Hilbert\ScrollArea, #PB_ScrollArea_InnerHeight, *Object_Hilbert\Square_Size)
      ResizeGadget(*Object_Hilbert\Canvas, 0, 0, *Object_Hilbert\Square_Size, *Object_Hilbert\Square_Size)
    EndIf
    
    *Object_Hilbert\Square_Size = *Object_Hilbert\Width
    
    If *Object_Hilbert\Square_Size     >= 1024 : Iterations = 10
    ElseIf *Object_Hilbert\Square_Size >= 512  : Iterations = 9
    ElseIf *Object_Hilbert\Square_Size >= 256  : Iterations = 8
    ElseIf *Object_Hilbert\Square_Size >= 128  : Iterations = 7
    ElseIf *Object_Hilbert\Square_Size >=  64  : Iterations = 6
    Else : ProcedureReturn #False;                      : Iterations = 5
    EndIf
    
    If Not StartDrawing(CanvasOutput(*Object_Hilbert\Canvas))
      ProcedureReturn #False
    EndIf
    
    Box(0, 0, *Object_Hilbert\Width, *Object_Hilbert\Height, RGB(255,255,255))
    
    Hilbert_State.Hilbert_State
    Hilbert_State\X = 0
    Hilbert_State\Y = 0
    Hilbert_State\Offset = 1
    Hilbert_State\Step_Size = Int_Divide_RUP(*Object_Hilbert\Object\Total_Size, (1<<Iterations)*(1<<Iterations))
    If *Object_Hilbert\Zoom_Factor > 0
      Hilbert_State\Step_Size >> *Object_Hilbert\Zoom_Factor
    ElseIf *Object_Hilbert\Zoom_Factor < 0
      Hilbert_State\Step_Size << -*Object_Hilbert\Zoom_Factor
    EndIf
    If Hilbert_State\Step_Size < 1 : Hilbert_State\Step_Size = 1 : EndIf
    Hilbert_State\Size_X = *Object_Hilbert\Square_Size
    Hilbert_State\Size_Y = *Object_Hilbert\Square_Size
    Hilbert_State\Object = *Object_Hilbert\Object
    
    Data_Source_Get_Ranges(*Object_Hilbert\Object\Data_Source, *Object_Hilbert\Object)
    ResetList(Hilbert_State\Object\Range())
    Hilbert_Level(Iterations, Hilbert_State)
    
    DrawingMode(#PB_2DDrawing_Transparent)
    DrawingFont(FontID(Main\Font_ID))
    
    
    ForEach *Object_Hilbert\Object\Range()
      If *Object_Hilbert\Object\Range()\Type_State & #MEM_FREE
        ; #### Memory is free
        Mem_Free.q + *Object_Hilbert\Object\Range()\Size
      ElseIf *Object_Hilbert\Object\Range()\Type_State & #MEM_COMMIT
        If *Object_Hilbert\Object\Range()\Type_State & #MEM_IMAGE
          ; #### Memory is committed (Image)
          Mem_Commit_Image.q + *Object_Hilbert\Object\Range()\Size
        ElseIf *Object_Hilbert\Object\Range()\Type_State & #MEM_MAPPED
          ; #### Memory is committed (Mapped)
          Mem_Commit_Mapped.q + *Object_Hilbert\Object\Range()\Size
        ElseIf *Object_Hilbert\Object\Range()\Type_State & #MEM_PRIVATE
          ; #### Memory is committed (Private)
          Mem_Commit_Private.q + *Object_Hilbert\Object\Range()\Size
        EndIf
      ElseIf *Object_Hilbert\Object\Range()\Type_State & #MEM_RESERVE
        ; #### Memory is reserved #### There are also (Image, Mapped, Private), but i ignore them here
        Mem_Reserved.q + *Object_Hilbert\Object\Range()\Size
      EndIf
    Next
    
    ix = (1<<Iterations)+10
    iy = 10
    Box(ix, iy, 200, 20 + 15 * Main\Font_Height, RGB(127, 127, 127))
    ix + 10
    iy + 10
    DrawText(ix, iy, "Free:", #Object_Hilbert_Color_Free)
    iy + Main\Font_Height
    DrawText(ix, iy, StrF(Mem_Free/1000000, 3)+" MB", #Object_Hilbert_Color_Free)
    iy + Main\Font_Height
    DrawText(ix, iy, "Committed (IMAGE):", #Object_Hilbert_Color_Commit_Image)
    iy + Main\Font_Height
    DrawText(ix, iy, StrF(Mem_Commit_Image/1000000, 3)+" MB", #Object_Hilbert_Color_Commit_Image)
    iy + Main\Font_Height
    DrawText(ix, iy, "Committed (MAPPED):", #Object_Hilbert_Color_Commit_Mapped)
    iy + Main\Font_Height
    DrawText(ix, iy, StrF(Mem_Commit_Mapped/1000000, 3)+" MB", #Object_Hilbert_Color_Commit_Mapped)
    iy + Main\Font_Height
    DrawText(ix, iy, "Committed (PRIVATE):", #Object_Hilbert_Color_Commit_Private)
    iy + Main\Font_Height
    DrawText(ix, iy, StrF(Mem_Commit_Private/1000000, 3)+" MB", #Object_Hilbert_Color_Commit_Private)
    iy + Main\Font_Height
    DrawText(ix, iy, "Reserved:", #Object_Hilbert_Color_Reserved)
    iy + Main\Font_Height
    DrawText(ix, iy, StrF(Mem_Reserved/1000000, 3)+" MB", #Object_Hilbert_Color_Reserved)
    iy + Main\Font_Height * 2
    DrawText(ix, iy, "Total:", RGB(255,255,255))
    iy + Main\Font_Height
    DrawText(ix, iy, StrF(*Object_Hilbert\Object\Total_Size/1000000, 3)+" MB", RGB(255,255,255))
    iy + Main\Font_Height * 2
    DrawText(ix, iy, "Zoom: "+Str(*Object_Hilbert\Zoom_Factor), RGB(255,255,255))
    
    StopDrawing()
  EndIf
EndProcedure

Procedure Object_Hilbert_Destroy(*Object_Hilbert.Object_Hilbert)
  If Not *Object_Hilbert
    ProcedureReturn #False
  EndIf
  
  *Object_Hilbert\Object\Hilbert = 0
  
  If *Object_Hilbert\Window_ID
    CloseWindow(*Object_Hilbert\Window_ID)
  EndIf
  
  ClearStructure(*Object_Hilbert, Object_Hilbert)
  FreeMemory(*Object_Hilbert)
  
  Window_Main_Panel_Rebuild(0)
  
  ProcedureReturn #True
EndProcedure

Procedure Object_Hilbert_Create(*Object.Object)
  If Not *Object
    ProcedureReturn #False
  EndIf
  
  Protected *Object_Hilbert.Object_Hilbert = AllocateMemory(SizeOf(Object_Hilbert))
  If Not *Object_Hilbert
    ProcedureReturn #False
  EndIf
  InitializeStructure(*Object_Hilbert, Object_Hilbert)
  
  *Object_Hilbert\Object = *Object
  
  *Object_Hilbert\Window_ID = AddGadgetItem(Window_Main\MDI, #PB_Any, "Hilbert: "+*Object\Name, 0, #PB_Window_Maximize | #PB_Window_MaximizeGadget | #PB_Window_MinimizeGadget | #PB_Window_SystemMenu | #PB_Window_SizeGadget)
  If Not *Object_Hilbert\Window_ID
    Object_Hilbert_Destroy(*Object_Hilbert)
    ProcedureReturn #False
  EndIf
  
  SmartWindowRefresh(*Object_Hilbert\Window_ID, 1)
  
  *Object_Hilbert\Width = 0
  *Object_Hilbert\Height = 0
  
  *Object_Hilbert\ScrollArea = ScrollAreaGadget(#PB_Any, 0, 0, *Object_Hilbert\Width, *Object_Hilbert\height, *Object_Hilbert\Width, *Object_Hilbert\height, 100)
  *Object_Hilbert\Canvas = CanvasGadget(#PB_Any, 0, 0, *Object_Hilbert\Width, *Object_Hilbert\Height, #PB_Canvas_Keyboard)
  
  ProcedureReturn *Object_Hilbert
EndProcedure

Procedure Object_Hilbert_Event(*Object_Hilbert.Object_Hilbert, Event, Event_Type)
  Select Event
    Case #PB_Event_ActivateWindow
      Window_Main_Panel_Rebuild(*Object_Hilbert\Window_ID)
      
    Case #PB_Event_SizeWindow, #PB_Event_MaximizeWindow, #PB_Event_RestoreWindow, #PB_Event_MinimizeWindow, #PB_Event_Repaint
      ;Object_Hilbert_View_Redo(*Object_Hilbert, )
      *Object_Hilbert\Redo = 1
      
    Case #PB_Event_Gadget
      Select EventGadget()
        Case *Object_Hilbert\Canvas
          Select Event_Type
            Case #PB_EventType_MouseWheel
              *Object_Hilbert\Zoom_Factor + GetGadgetAttribute(*Object_Hilbert\Canvas, #PB_Canvas_WheelDelta)
              *Object_Hilbert\Redo = 2
              
          EndSelect
      EndSelect
      
    Case #PB_Event_CloseWindow
      Object_Hilbert_Destroy(*Object_Hilbert)
      
  EndSelect
EndProcedure

Procedure Object_Hilbert_Main(*Object_Hilbert.Object_Hilbert)
  If *Object_Hilbert\Redo Or *Object_Hilbert\Window_ID = GetGadgetState(Window_Main\MDI)
    *Object_Hilbert\Redo = 0
    
    If *Object_Hilbert\Redo = 2 Or (*Object_Hilbert\Redo_Timer < ElapsedMilliseconds() And *Object_Hilbert\Window_ID = GetGadgetState(Window_Main\MDI))
      *Object_Hilbert\Redo_Timer = ElapsedMilliseconds() + 100
      
      Object_Hilbert_View_Redo(*Object_Hilbert, WindowWidth(*Object_Hilbert\Window_ID), WindowHeight(*Object_Hilbert\Window_ID))
    EndIf
  EndIf
EndProcedure
; IDE Options = PureBasic 4.60 RC 2 (Windows - x64)
; CursorPosition = 338
; FirstLine = 336
; Folding = --
; EnableXP
; DisableDebugger